class ApiController < ApplicationController

  TRANSFER_FILTER_FIELDS = {
      'contact' => {
          :forbidden_comparators => [],
          :generic_type => :String,
          :map_to => 'contact',
          :table => 'jobs'
      },
      'filename' => {
          :forbidden_comparators => ['contain', 'match_re'],
          :generic_type => :String,
          :map_to => 'file_basename',
          :table => 'fasp_files'
      },
      'started_at' => {
          :forbidden_comparators => ['start_with', 'contain', 'match_re'],
          :generic_type => :Datetime,
          :map_to => 'started_at',
          :table => 'job_steps'
      },
      'status' => {
          :forbidden_comparators => [],
          :generic_type => :String,
          :map_to => 'status',
          :table => 'job_steps'
      },
      'stopped_at' => {
          :forbidden_comparators => ['start_with', 'contain', 'match_re'],
          :generic_type => :Datetime,
          :map_to => 'ended_at',
          :table => 'job_steps'
      },
      'transfer_name' => {
          :forbidden_comparators => [],
          :generic_type => :String,
          :map_to => 'description',
          :table => 'jobs'
      }
  }

  TRANSFER_FILTER_COMPARATORS = [
      'contain',
      'eq',
      'gt',
      'gte',
      'in',
      'is_not_null',
      'is_null',
      'lt',
      'lte',
      'match_re',
      'neq',
      'start_with'
  ]

  # equivalent to submitting a simple transfer
  def create
    # convert from API request format to parameter format needed by TransferJob
    transfer_params = {}
    email_params = {}
    tags_params = {}
    transfer_params[:description] = params[:name]
    transfer_params[:comment] = params[:comment] || ""

    transfer_params[:source_endpoint_type] = params[:source][:endpoint_type]
    transfer_params[:source_end_point_id] = params[:source][:endpoint_id]
    transfer_params[:source_login] = params[:source][:user]
    transfer_params[:source_address] = params[:source][:ip_address]
    transfer_params[:source_port] = params[:source][:port]
    transfer_params[:source_use_ssh_key] = params[:source][:authentication_method] == 'ssh_key'
    transfer_params[:source_password] = params[:source][:password]
    transfer_params[:source_ssh_key_id] = params[:source][:ssh_key_id]
    transfer_params[:source_use_ssl] = params[:source][:use_ssl]
    transfer_params[:source_validate_ssl_cert] = params[:source][:validate_ssl_cert]
    transfer_params[:source_paths] = params[:source][:paths].join("\n")
    transfer_params[:source_base_enable] = params[:source][:source_base].present?
    transfer_params[:source_base] = params[:source][:source_base]

    transfer_params[:dest_endpoint_type] = params[:destination][:endpoint_type]
    transfer_params[:dest_end_point_id] = params[:destination][:endpoint_id]
    transfer_params[:dest_login] = params[:destination][:user]
    transfer_params[:dest_address] = params[:destination][:ip_address]
    transfer_params[:dest_port] = params[:destination][:port]
    transfer_params[:dest_use_ssh_key] = params[:destination][:authentication_method] == 'ssh_key'
    transfer_params[:dest_password] = params[:destination][:password]
    transfer_params[:dest_use_ssl] = params[:destination][:use_ssl]
    transfer_params[:dest_validate_ssl_cert] = params[:destination][:validate_ssl_cert]
    transfer_params[:dest_ssh_key_id] = params[:destination][:ssh_key_id]
    transfer_params[:dest_path] = params[:destination][:path]

    transfer_params[:udp_port] = params[:udp_port]

    transfer_params[:target_rate] = params[:target_rate_kbps]
    transfer_params[:minimum_rate] = params[:minimum_rate_kbps]
    transfer_params[:priority_and_policy] = params[:policy]

    transfer_params[:enable_retry] = params[:retries][:enable] if params[:retries].present? && params[:retries].has_key?(:enable)
    transfer_params[:max_retry_attempts] = params[:retries][:attempts] if params[:retries].present? && params[:retries].has_key?(:attempts)
    transfer_params[:retry_interval_seconds] = params[:retries][:interval_seconds] if params[:retries].present? && params[:retries].has_key?(:interval_seconds)

    transfer_params[:file_exclude_by_timestamp] = params[:file_exclude_by_timestamp] || false
    transfer_params[:file_exclude_newer_than] = params[:file_exclude_newer_than] if params[:file_exclude_by_timestamp]

    transfer_params[:encrypt_at_rest_enable] = params[:content_protection][:enable]
    transfer_params[:encryption_password] = params[:content_protection][:password]
    transfer_params[:encryption_password_confirm] = params[:content_protection][:password]
    transfer_params[:encryption] = params[:transport_encryption]

    transfer_params[:resume_incomplete] = params[:resume][:enable]
    transfer_params[:file_compare_type] = params[:resume][:comparison_method]
    transfer_params[:file_overwrite_policy] = params[:overwrite_policy]
    transfer_params[:preserve_file_timestamps] = params[:preserve_file_timestamps]

    if params[:file_attributes].present?
      transfer_params[:preserve_file_uids] = params[:file_attributes][:preserve_file_uids]
      transfer_params[:preserve_file_gids] = params[:file_attributes][:preserve_file_gids]
      transfer_params[:preserve_file_timestamps] = params[:file_attributes][:preserve_file_timestamps]
      if params[:file_attributes][:preserve_file_timestamps].present? && params[:file_attributes][:preserve_file_timestamps]
        transfer_params[:preserve_file_create_times] = params[:file_attributes][:preserve_file_create_times]
        transfer_params[:preserve_file_modify_times] = params[:file_attributes][:preserve_file_modify_times]
        transfer_params[:preserve_file_access_times] = params[:file_attributes][:preserve_file_access_times]
        transfer_params[:preserve_file_source_access_times] = params[:file_attributes][:preserve_file_source_access_times]
      end
    end
    
    transfer_params[:symbolic_link_handling] = params[:symbolic_links]
    transfer_params[:archive_after_transfer] = params[:archive_after_transfer].present? && params[:archive_after_transfer][:enable] || false
    transfer_params[:archive_directory] = params[:archive_directory]
    transfer_params[:delete_source_files] = params[:delete_source_files_after_transfer].present? && params[:delete_source_files_after_transfer][:enable] || false
    transfer_params[:delete_empty_directories] = params[:delete_source_files_after_transfer].present? && params[:delete_source_files_after_transfer][:also_delete_empty_directories] || false
    transfer_params[:exclude_file_filter] = params[:exclude_file_filter]
    transfer_params[:include_file_filter] = params[:include_file_filter]

    transfer_params[:datagram_size_cbx] = params[:mtu_size].present?
    transfer_params[:datagram_size] = params[:mtu_size]

    transfer_params[:read_block_size_cbx] = params[:read_block_size].present?
    transfer_params[:read_block_size] = params[:read_block_size]

    transfer_params[:write_block_size_cbx] = params[:write_block_size].present?
    transfer_params[:write_block_size] = params[:write_block_size]

    transfer_params[:run_when] = params[:scheduled_start].present? ? 'later' : 'now'
    transfer_params[:scheduled_start_at] = params[:scheduled_start]

    if params[:fasp_proxy].present? && params[:fasp_proxy][:enable]
      transfer_params[:use_fasp_proxy] = true

      transfer_params[:proxy_host] = params[:fasp_proxy][:host]
      transfer_params[:proxy_port] = params[:fasp_proxy][:port]
      transfer_params[:proxy_secure] = params[:fasp_proxy][:use_secure_dnat]
      transfer_params[:proxy_user] = params[:fasp_proxy][:login] if params[:fasp_proxy][:login].present?
      transfer_params[:proxy_password] = params[:fasp_proxy][:password] if params[:fasp_proxy][:password].present?
    end

    if params[:notifications].present?
      fake_email_id = 1
      params[:notifications].each do |notification|
        email_params[fake_email_id] = {
            :email => notification[:email_address],
            :id => fake_email_id,
            :start_template_id => notification[:start_template_id],
            :success_template_id => notification[:success_template_id],
            :error_template_id => notification[:error_template_id]
        }
        fake_email_id += 1
      end
    end

    if params[:tags].present?
      tag_id = 1
      params[:tags].each do |tag|
        tags_params[tag_id] = {
            :name => tag[:name],
            :value => tag[:value]
        }
        tag_id += 1
      end
      transfer_params[:console_tags] = tags_params.to_json
    end
    
    # if source or dest end point id is NULL, we need to substitute in the appropriate wild card endpoints (if the user has permissions for them)
    if transfer_params[:source_end_point_id].blank?
      # does user have permission to do 'Any'?
      universal_endpoint = current_user.source_logins_for_start_simple.find {|endpoint|
        endpoint.login == '*' && endpoint.address == '*'
      }
      if universal_endpoint
        transfer_params[:source_end_point_id] = universal_endpoint.id
      else
        render :json => {"error" => "no permission to enter arbitrary source login info"}, :status => 403
        return
      end
    end

    if transfer_params[:dest_end_point_id].blank?
      # does user have permission to do 'Any'?
      universal_endpoint = current_user.dest_logins_for_start_simple(Login.find(transfer_params[:source_end_point_id])).find {|endpoint|
        endpoint.login == '*' && endpoint.address == '*'
      }
      if universal_endpoint
        transfer_params[:dest_end_point_id] = universal_endpoint.id
      else
        render :json => {"error" => "no permission to enter arbitrary dest login info for the specified source"}, :status => 403
        return
      end
    end

    @transfer = TransferJob.new(current_user, transfer_params, email_params)

    if @transfer.save
      render :json => {:id => @transfer.step.id}, :status => 200
    else
      raise "failed to save transfer #{@transfer.errors.full_messages}"
    end
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  # User's available endpoints
  def endpoints
    merged_endpoints = {}
    source_endpoints = current_user.source_logins_for_start_simple
    dest_endpoints = current_user.dest_logins_for_start_simple
    source_endpoints.each {|endpoint|
      merged_endpoints[endpoint.id] = {}
      merged_endpoints[endpoint.id][:id] = endpoint.id
      merged_endpoints[endpoint.id][:credential_type] = endpoint.ssh_key_id.nil? ? (endpoint.has_password? ? 'password' : 'none') : 'ssh_key'
      merged_endpoints[endpoint.id][:ssh_key_id] = endpoint.ssh_key_id
      merged_endpoints[endpoint.id][:ip_address] = endpoint.address
      merged_endpoints[endpoint.id][:ssh_port] = endpoint.port || 22
      merged_endpoints[endpoint.id][:login] = endpoint.login

      merged_endpoints[endpoint.id][:endpoint_type] = endpoint.owner_type == 'User' ? 'Personal' : 'Node'
      merged_endpoints[endpoint.id][:description] = endpoint.name

      merged_endpoints[endpoint.id][:source] = true
      merged_endpoints[endpoint.id][:dest] = false
    }

    dest_endpoints.each {|endpoint|
      merged_endpoints[endpoint.id] ||= {}
      merged_endpoints[endpoint.id][:id] = endpoint.id
      merged_endpoints[endpoint.id][:credential_type] = endpoint.ssh_key_id.nil? ? (endpoint.has_password? ? 'password' : nil) : 'ssh_key'
      merged_endpoints[endpoint.id][:ssh_key_id] = endpoint.ssh_key_id
      merged_endpoints[endpoint.id][:ip_address] = endpoint.address
      merged_endpoints[endpoint.id][:ssh_port] = endpoint.port || 22
      merged_endpoints[endpoint.id][:login] = endpoint.login

      merged_endpoints[endpoint.id][:endpoint_type] = endpoint.owner_type == 'User' ? 'Personal' : 'Node'
      merged_endpoints[endpoint.id][:description] = endpoint.name

      merged_endpoints[endpoint.id][:source] ||= false
      merged_endpoints[endpoint.id][:dest] = true
    }

    result = []
    merged_endpoints.each_value {|v|
      result << v
    }
    result.sort! {|a,b| a[:id] <=> b[:id]}

    render :json => result, :status => 200
  end

  def ssh_keys
    result = current_user.ssh_keys.collect {|key|
      {
          :id => key.id,
          :name => key.name,
          :filename_on_node => key.basename
      }
    }
    render :json => result, :status => 200
  end

  # Retrieve list of transfers (each returned item is a JobStep)
  def index
    from_raw = params[:from]
    to_raw = params[:to]

    from = nil
    to = nil

    Time.use_zone('UTC') {
      from = Time.zone.parse(from_raw)
      to = to_raw.present? ? Time.zone.parse(to_raw) : nil
    }

    # check for failed parse or missing from
    if from.nil?
      error_message = "'from' parameter missing or not a valid date/time"
      raise ArgumentError, error_message
    end

    if to_raw.present? && to.nil?
      error_message = "'to' parameter not a valid date/time"
      raise ArgumentError, error_message
    end

    to ||= Time.now.utc


    if to - from > 24.hours
      error_message = "More than 24 hours data requested"
      raise ArgumentError, error_message
    end

    if to < from
      error_message = "'From' date must be before 'to' date"
      raise ArgumentError, error_message
    end

    # process additional filters
    additional_filters = []
    join_jobs_table = false
    filter_index = 1
    while params["filter#{filter_index}"].present?
      field = params["filter#{filter_index}"]
      comparator = params["comp#{filter_index}"]
      value = params["val#{filter_index}"]

      error_message ||= "bad field for filter #{filter_index}" unless TRANSFER_FILTER_FIELDS.keys.include?(field)
      error_message ||= "missing comparator for filter #{filter_index}" unless comparator.present?
      error_message ||= "bad comparator for filter #{filter_index}" unless TRANSFER_FILTER_COMPARATORS.include?(comparator)
      error_message ||= "comparator not allowed for field for filter #{filter_index}" if TRANSFER_FILTER_FIELDS[field][:forbidden_comparators].include?(comparator)
      raise ArgumentError, error_message if error_message

      # -----------------------------------
      # convert filter to SQL expression
      # -----------------------------------

      operator = AspExp::Operator[comparator.to_sym]
      # process value(s) for this filter
      args = []
      unless operator.unary?
        # split values for "IN" operator by comma
        if operator.name == :in
          args += value.split(',').map(& :strip)
        else
          args << value
        end
      end

      generic_type = TRANSFER_FILTER_FIELDS[field][:generic_type]

      # convert args to proper types
      processed_args = args.collect { |arg|
        AspExp::DataType.coerce_to_generic_type(arg, generic_type)
      }

      # map to field name, add special handling if in different table than job_steps, get sql expression for condition
      remapped_field = "#{TRANSFER_FILTER_FIELDS[field][:table]}.#{TRANSFER_FILTER_FIELDS[field][:map_to]}"
      if TRANSFER_FILTER_FIELDS[field][:table] == 'jobs'
        join_jobs_table = true
      end
      filter_sql = operator.filter_sql_expression([remapped_field, processed_args].flatten)
      # If this is a filter on filename, then wrap in an EXISTS clause
      if TRANSFER_FILTER_FIELDS[field][:table] == 'fasp_files'
        filter_sql = JobStep.api_file_filter_clause(filter_sql)
      end

      additional_filters << filter_sql

      filter_index += 1
    end

    # add SQL expression for date range filter
    all_filters = additional_filters +
        [JobStep.public_sanitize_sql_for_conditions(["job_steps.started_at <= ? AND (job_steps.ended_at >= ? OR job_steps.ended_at IS NULL)", to, from])]

    # join filters together with AND
    filter_sql = "(#{all_filters.join(') AND (')})"

    # get job_steps that have a 'started_at'
    # TODO: performance optimization - may need to supply index hint to use idx_js_started_at
    job_steps = current_user.job_steps.find(:all, :conditions => filter_sql, :joins => (join_jobs_table ? "INNER JOIN jobs ON job_steps.job_id = jobs.id" : nil))

    # get job_steps that don't have a 'started_at' yet, e.g. queued, scheduled, and submitted job_steps
    # TODO: performance optimization - may need to supply index hint
    all_filters = additional_filters +
        [
            "job_steps.started_at IS NULL",
            JobStep.public_sanitize_sql_for_conditions(
                ["(job_steps.scheduled_start_at BETWEEN ? AND ?) OR (job_steps.submitted_at BETWEEN ? AND ?) OR (job_steps.queued_at BETWEEN ? AND ?)",
                 from, to, from, to, from, to]),
        ]
    filter_sql = "(#{all_filters.join(') AND (')})"

    job_steps_without_start_dates = current_user.job_steps.find(:all, :conditions => filter_sql, :joins => (join_jobs_table ? "INNER JOIN jobs ON job_steps.job_id = jobs.id" : nil))

    json_hashes = []

    (job_steps + job_steps_without_start_dates).uniq.each do |job_step|
      json_hashes << transfer_hash(job_step)
    end

    render :json => json_hashes, :status => 200
  rescue
    logger.error "#{$!} #{$@.join("\n")}"
    render :json => {:error => error_message || "Exception raised"}, :status => 422
  end

  def show
    step = current_user.job_steps.find(params[:id]) rescue nil
    if step.nil?
      status_code = 422 # unprocessable entity
      response_hash = nil
    else
      status_code = 200 # ok
      response_hash = transfer_hash(step, true)
    end
    render :json => response_hash, :status => status_code
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  # start a scheduled transfer immediately
  def start
    request_transfer_action(:start)
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def cancel
    request_transfer_action(:cancel)
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def pause
    request_transfer_action(:pause)
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def resume
    request_transfer_action(:resume)
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def rerun
    request_transfer_action(:rerun)
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def change_rate
    min_rate = params[:min_rate_kbps]
    target_rate = params[:target_rate_kbps]

    # check valid numbers
    if min_rate.to_i.to_s != min_rate
      raise "Bad minimum rate"
    end

    if target_rate.to_i.to_s != target_rate
      raise "Bad target rate"
    end

    raise "Bad rates #{min_rate},#{target_rate}" unless min_rate.present? && target_rate.present? && min_rate.to_i >= 0 &&
        target_rate.to_i >= 0 && min_rate.to_i <= target_rate.to_i
    step = current_user.job_steps.find(params[:id])
    if current_user.can_control_step?(step)
      min_rate_kbps = params[:min_rate_kbps].to_i
      target_rate_kbps = params[:target_rate_kbps].to_i
      policy = step.session.last_rate.policy
      step.session.request_rate_change(min_rate_kbps, target_rate_kbps, (policy=="Adaptive" ? "Fair" : policy))
    else
      render :json => nil, :status => 403
      return
    end
    render :json => nil, :status => 200
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def change_policy
    policy = params[:policy]
    raise "Bad policy #{policy}" unless ['Fair', 'Fixed', 'Trickle'].include?(policy)
    step = current_user.job_steps.find(params[:id])
    if current_user.can_control_step?(step)
      step.session.request_policy_change(policy)
    else
      render :json => nil, :status => 403
      return
    end
    render :json => nil, :status => 200
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def queue_contents
    queue_id = params[:queue_id]

    queue = AspQ::Queue.find(queue_id)

    # check that user has permission to see queue
    node =  queue.resource
    if node.is_a?(FaspNode)
      unless (current_user.nodes.find(node.id) rescue nil)
        render :json => nil, :status => 403
        return
      end
    else
      unless (current_user.partner_nodes.find(node.id) rescue nil)
        render :json => nil, :status => 403
        return
      end
    end

    response = []
    queue.spots.each do |spot|
      response << {:item_id => spot.item_id, :waiting => spot.waiting}
    end
    render :json => response, :status => 200
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  # promote a transfer on a queue
  def move_forwards
    queue_id = params[:queue_id]
    unless AspQ::Queue.find(queue_id).reorderable_by?(current_user)
      render :json => nil, :status => 403
      return
    end
    job_step_id = params[:id]
    spot = AspQ::QueueSpot.find_by_queue_id_and_item_id(queue_id, job_step_id)
    spot.promote
    render :json => nil, :status => 200
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  # demote a transfer on a queue
  def move_back
    queue_id = params[:queue_id]
    unless AspQ::Queue.find(queue_id).reorderable_by?(current_user)
      render :json => nil, :status => 403
      return
    end
    job_step_id = params[:id]
    spot = AspQ::QueueSpot.find_by_queue_id_and_item_id(queue_id, job_step_id)
    spot.demote
    render :json => nil, :status => 200
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def transfer_file_list
    offset = params[:offset].to_i
    limit = params[:limit].to_i
    path_filter = params[:path_match]
    errors_only = params[:errors_only] == "true"

    status_code = 422 # default to failure: unprocessable entity

    if limit > 100
      response = {"error" => "More than 100 items requested"}
    elsif limit <= 0
      response = {"error" => "Bad limit parameter"}
    elsif offset < 0
      response = {"error" => "Bad limit parameter"}
    else
      step = current_user.job_steps.find(params[:id]) rescue nil
      if step.nil?
        response = {"error" => "Transfer not found, or you do not have permission to see it"}
      elsif step.session.nil?
        response = {"error" => "No session info for transfer yet"}
      else
        # get files
        where_clause = step.file_search_sql(errors_only, path_filter)
        sql = "SELECT * FROM fasp_files USE INDEX (index_fasp_files_on_job_step_id) WHERE #{where_clause} ORDER BY id DESC LIMIT #{offset}, #{limit}"
        files = FaspFile.find_by_sql(sql)
        response = []
        files.each do |file|
          response << {
              :bytes_transferred => file.calc_bytes_transferred,
              :bytes_written => file.bytes_written,
              :error_description => file.err_desc,
              :id => file.id,
              :path => file.file_fullpath,
              :size => file.size,
              :started_at => file.started_at,
              :status => file.status,
              :stopped_at => file.stopped_at,
              :checksum => file.checksum,
              :checksum_type => file.checksum_type
          }
        end
        status_code = 200 # ok
      end
    end
    render :json => response, :status => status_code
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end


  # list of smart transfers
  def smart_transfers
    transfers = current_user.job_templates.find_all{|template| template.startable_by?(current_user)}
    result_hashes = []
    transfers.each do |transfer|
      result_hashes << {
          :id => transfer.id,
          :name => transfer.description,
          :owner => transfer.user && transfer.user.login
      }
    end
    render :json => result_hashes, :status => 200
  end

  def smart_transfer_submit
    smart_transfer = current_user.job_templates.find(params[:id]) rescue nil

    if smart_transfer.nil?
      render :json => nil, :status => 404
      return
    end

    unless smart_transfer.startable_by?(current_user)
      render :json => nil, :status => 403
      return
    end

    transfer_params = {}
    transfer_params[:description] = params[:name]
    transfer_params[:comment] = params[:comment] || ""
    transfer_params[:source_paths] = params[:source][:paths].join("\n") rescue nil
    transfer_params[:source_base_enable] = params[:source][:source_base].present? rescue nil
    transfer_params[:source_base] = params[:source][:source_base] rescue nil
    transfer_params[:scheduled_start] = params[:scheduled_start]

    new_job = smart_transfer.submit_from_api_call(current_user, transfer_params)

    unless new_job
      raise smart_transfer.errors.full_messages.join("\n")
    end

    result_array = []
    new_job.steps.each do |step|
      result_array << {:id => step.id}
    end
    render :json => result_array, :status => 200
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

  def email_server_update
    unless current_user.admin
      render :json => {:error => "Only Admins can update Email Server Details"}, :status => 422
      return
    end
    email_server_params = {}

    email_server_params[:server] = params[:server] if params[:server].present?
    email_server_params[:port] = params[:port] if params[:port].present?
    email_server_params[:domain] = params[:domain] if params[:domain].present?
    email_server_params[:tls] =  params[:use_tls] == true ? 1 : 0 if params[:use_tls].present?
    if params[:authentication].present?
      if params[:authentication][:type] == "login"
        email_server_params[:authentication] = params[:authentication][:type]
        email_server_params[:user_name] = params[:authentication][:user_name] if params[:authentication][:user_name].present?
        email_server_params[:password] = params[:authentication][:password] if params[:authentication][:password].present?
      elsif params[:authentication][:type] == "none"
        email_server_params[:authentication] = "none"
      else
        logger.error "#{$!} #{$@}"
        render :json => {:error => "Error : Authentication Type Required"}, :status => 422
        return
      end
    end
    email_server_params[:from_address] = params[:from_address] if params[:from_address].present?
    email_server_params[:from_name] = params[:from_name] if params[:from_name].present?

    @email_server = ConsoleConfig::EmailServer.load_settings
    @email_server.attributes = email_server_params

    if @email_server.save_settings
      render :json => 'Email Server settings saved.', :status => 200
    end
    rescue Exception => e
      logger.error "#{$!} #{$@}"
      render :json => {:error => "Exception raised : #{e.message}"}, :status => 422
  end

  def nodeapi_credentials_update
    unless current_user.admin
      render :json => {:error => "Only Admins can update Credentials"}, :status => 422
      return
    end
    if !params[:login].present? || !params[:password].present?
      render :json => {:error => "User name and Password Required"}, :status => 422
      return
    end
    @node = FaspNode.find_by_node_address(params[:node_address])
    if @node.nil?
      render :json => {:error => "Node not found in Console, Enter valid Node Address"}, :status => 422
      return
    end
    if params[:auth_type].present? && (params[:auth_type] == "NodeAPI" || params[:auth_type] == "SSH")
      if params[:auth_type] == "NodeAPI"
        begin
          @node.test_node_api_credentials(params[:login], params[:password], @node.v3_node_daemon_port, @node.v3_node_use_ssl, @node.v3_validate_ssl_cert)
          @node.node_api_username = params[:login]
          @node.node_api_password = params[:password]
        rescue Exception => e
          render :json => {:error => "Failed to connect via NodeAPI: #{e.message}"}, :status => 422
          return
        end
        if @node.update_node_api_login
          render :json => 'Node API Credentials updated.', :status => 200
        else
          render :json => {:error => "Failed to update NodeAPI password"}, :status => 422
          return
        end
      elsif params[:auth_type] == "SSH"
        begin
          @node.test_config_credentials(params[:login], params[:password], nil)
          @node.configuration_login = params[:login]
          @node.configuration_password = params[:password]
            # if we successfully connected, try to get asws info
            #query_node
        rescue Exception => e
          render :json => {:error => "Failed to connect via SSH: #{e.message}"}, :status => 422
          return
        end
        if @node.update_config_login
          render :json => 'SSH Credentials updated.', :status => 200
        else
          render :json => {:error => "Failed to update SSH password"}, :status => 422
          return
        end
      end
    else
      render :json => {:error => "Error : Valid Authentication Type Required"}, :status => 422
      return
    end
  rescue Exception => e
    logger.error "#{$!} #{$@}"
    render :json => {:error => "Exception raised : #{e.message}"}, :status => 422
  end

  #========================================================================
  protected
  #========================================================================

  # override generic redirect_to_login, return an auth failure instead
  def redirect_to_login(was_saml_user=false)
    render :json => nil, :status => 401 # unauthorized
  end




  #========================================================================
  private
  #========================================================================


  def transfer_hash(step, detailed = false)
    result = {
        :bytes_transferred => step.total_bytes_transferred,
        :bytes_written => step.session && step.session.bytes_written,
        :contact => step.job.contact,
        :destination => step.dest_node_address,
        :destination_description => step.dest_node_name,
        :error_description => step.err_desc,
        :eta => step.eta && step.eta.utc,
        :id => step.id,
        :name => step.job.description,
        :queued_at => step.queued_at && step.queued_at.utc,
        :source => step.source_node_address,
        :source_description => step.source_node_name,
        :scheduled_start_at => step.scheduled_start_at && step.scheduled_start_at.utc,
        :started_at => step.started_at && step.started_at.utc,
        :started_via => step.job.started_via,
        :status => step.status,
        :stopped_at => step.ended_at && step.ended_at.utc,
        :submitted_at => step.submitted_at && step.submitted_at.utc,
        :total_bytes => step.session && step.session.bytes_pretransfer,

    }
    if detailed

      queues = []
      step.queue_spots.each do |spot|
        queues << {
            :queue_id => spot.queue_id,
            :description => spot.queue.description,
            :position => spot.index,
            :waiting => spot.waiting
        }
      end

      result.merge!(
          {
              :comment => step.job.comment,
              :destination_description => step.dest_node_name,
              :destination_endpoint_id => step.dest_saved_login_id,
              :destination_path => step.dest_path,
              :initiator => step.source_initiated ? 'source' : 'destination',
              :min_rate_kbps => step.minimum_rate,
              :policy => step.priority_and_policy,
              :queues => queues,
              :source_description => step.source_node_name,
              :source_endpoint_id => step.source_saved_login_id,
              :source_paths => step.source_paths.to_s.split(/\n|\t/),
              :target_rate_kbps => step.target_rate,
              :transport_encryption => step.encryption

          }
      )

      session = step.session

      if session
        result.merge!(
            {
                :bytes_lost => session.bytes_lost,

                :files_completed => session.files_complete,
                :files_failed => session.files_failed,
                :files_skipped => session.files_skipped,

                :last_calculated_rate_bps => session.temp_bitrate,
                :min_rate_kbps => session.current_min_rate, # current rate may be different than what was initially set: override value from job_step
                :network_delay_ms => session.network_delay || 0,
                :policy => session.current_policy_and_priority,

                :session_uuid => session.session_id,
                :target_rate_kbps => session.current_target_rate, # current rate may be different than what was initially set: override value from job_step
                :total_files => session.files_pretransfer,
                :transport => session.transport,
                :transport_encryption => session.cipher
            }
        )
      else
        result.merge!(
            {
                :session_uuid => nil
            }
        )
      end
    end

    result
  end

  def api_request?
    true
  end

  def request_transfer_action(action)
    # see if step exists at all
    admin_step = JobStep.find(params[:id]) rescue nil
    unless admin_step
      render :json => nil, :status => 404
      return
    end

    # check to see if user can see step
    step = current_user.job_steps.find(params[:id]) rescue nil
    unless step
      render :json => nil, :status => 403
      return
    end

    unless current_user.can_control_step?(step)
      render :json => nil, :status => 403
      return
    end


    if action == :rerun
      unless step.console?
        raise "can't rerun a transfer not initiated by console"
      end
      return_val = step.request(current_user, action)
      # return id of new job_step
      render :json => {:id => return_val.id}, :status => 200
    else
      step.request(current_user, action)
      render :json => nil, :status => 200
    end
  rescue
    logger.error "#{$!} #{$@}"
    render :json => nil, :status => 422
  end

end
